import { Row } from './Row';
import { Column } from './Column';
import { CellDoesNotExistInGridError, InvalidGridSizeError } from './errors';
import { GridEventDispatcher } from './utils';
/**
 * The grid contains all information about cells
 */
export class Grid {
    options;
    initializeCell;
    width;
    height;
    _grid = [];
    _rows = [];
    _columns = [];
    eventDispatcher = new GridEventDispatcher();
    /**
     * @param options The initialization configuration
     * @param initializeCell This function will initialize a cell
     */
    constructor(options, initializeCell) {
        this.options = options;
        this.initializeCell = initializeCell;
        if ('initializeCellValue' in this.options) {
            this.width = this.options.width;
            this.height = this.options.height;
        }
        else if ('grid' in this.options) {
            const rows = this.options.grid;
            this.height = rows.length;
            this.width = Math.max(0, ...rows.map((row) => row.length));
        }
        else {
            throw `Invalid options provided`;
        }
    }
    initializeGrid() {
        if ('initializeCellValue' in this.options) {
            this._grid.push(...this.initializeCells(this.options.initializeCellValue));
        }
        if ('grid' in this.options) {
            const rows = this.options.grid;
            this._grid.push(...this.initializeCells(({ row, col }) => rows[row][col]));
        }
        for (let row = 0; row < this.height; row++) {
            this._rows.push(new Row(this, row));
        }
        for (let col = 0; col < this.width; col++) {
            this._columns.push(new Column(this, col));
        }
        const forwardEvent = (event) => this.eventDispatcher.dispatchCellValueChangedEvent(event);
        for (const cell of this.cells) {
            cell.onValueChanged(forwardEvent);
        }
    }
    initializeCells(initializeCellValue) {
        if (this.height === 0 || this.width === 0)
            throw new InvalidGridSizeError(this.width, this.height);
        const cells = new Array(this.height);
        for (let row = 0; row < this.height; row++) {
            cells[row] = new Array(this.width);
            for (let col = 0; col < this.width; col++) {
                const coordinate = { row, col };
                const value = initializeCellValue(coordinate);
                cells[row][col] = this.initializeCell(coordinate, value);
            }
        }
        return cells;
    }
    /**
     * @returns The 2-dimensional array representing a grid (1. dimension: row (y-axis), 2. dimension: column (x-axis))
     */
    get grid() {
        return this._grid;
    }
    /**
     * @param coordinate The coordinate of the cell
     * @returns true if the cell exists in the grid
     */
    cellExists(coordinate) {
        const { row, col } = coordinate;
        return col >= 0 && row >= 0 && row < this.height && col < this.width;
    }
    /**
     * @returns All cells in the grid in ascending order by row and column
     */
    get cells() {
        return this.grid.flat();
    }
    /**
     * @param coordinate The coordinate of the cell
     * @returns The cell
     * @throws {CellDoesNotExistInGridError} when the cell does not exist in the grid
     */
    getCell(coordinate) {
        if (!this.cellExists(coordinate))
            throw new CellDoesNotExistInGridError(this, coordinate);
        return this.grid[coordinate.row][coordinate.col];
    }
    /**
     * @param row The row coordinate
     * @returns The row
     */
    getRow(row) {
        return this._rows[row];
    }
    /**
     * @returns All rows of the grid in ascending order
     */
    get rows() {
        return this._rows;
    }
    /**
     * @param col The column coordinate
     * @returns The column
     */
    getColumn(col) {
        return this._columns[col];
    }
    /**
     * @returns All columns of the grid in ascending order
     */
    get columns() {
        return this._columns;
    }
    /**
     * @param callback A function that should be called, when a cell value of this grid changes
     * @returns a function to unregister the callback
     */
    onCellValueChanged(callback) {
        return this.eventDispatcher.onCellValueChanged(callback);
    }
}
