import type { Coordinate } from './Coordinate';
import { Row } from './Row';
import { Column } from './Column';
import type { Cell } from './Cell';
import { type CellValueChangedEvent } from './utils';
/**
 * Options to generate a new grid
 */
export interface InitializeGridOptions<Value> {
    /**
     * The amount of columns (min: 1)
     */
    width: number;
    /**
     * The amount of rows (min: 1)
     */
    height: number;
    /**
     * The given function will be called, when the cell at {coordinate} will be initialized. The returned value will be the value of the cell.
     * @param coordinate The coordinate of the cell that is being initialized
     */
    initializeCellValue: (coordinate: Coordinate) => Value;
}
/**
 * Options to generate a new grid based on an existing 2-dimensional array.
 */
export interface PreInitializedGridOptions<Value> {
    /**
     * The 2-dimensional array representing a grid (1. dimension: row (y-axis), 2. dimension: column (x-axis))
     */
    grid: Value[][];
}
/**
 * The grid contains all information about cells
 */
export declare class Grid<Value, CellWithValue extends Cell<Value>> {
    private options;
    private initializeCell;
    readonly width: number;
    readonly height: number;
    private readonly _grid;
    private readonly _rows;
    private readonly _columns;
    private readonly eventDispatcher;
    /**
     * @param options The initialization configuration
     * @param initializeCell This function will initialize a cell
     */
    constructor(options: InitializeGridOptions<Value> | PreInitializedGridOptions<Value>, initializeCell: (coordinate: Coordinate, value: Value) => CellWithValue);
    protected initializeGrid(): void;
    private initializeCells;
    /**
     * @returns The 2-dimensional array representing a grid (1. dimension: row (y-axis), 2. dimension: column (x-axis))
     */
    get grid(): CellWithValue[][];
    /**
     * @param coordinate The coordinate of the cell
     * @returns true if the cell exists in the grid
     */
    cellExists(coordinate: Coordinate): boolean;
    /**
     * @returns All cells in the grid in ascending order by row and column
     */
    get cells(): CellWithValue[];
    /**
     * @param coordinate The coordinate of the cell
     * @returns The cell
     * @throws {CellDoesNotExistInGridError} when the cell does not exist in the grid
     */
    getCell(coordinate: Coordinate): CellWithValue;
    /**
     * @param row The row coordinate
     * @returns The row
     */
    getRow(row: number): Row<Value, CellWithValue>;
    /**
     * @returns All rows of the grid in ascending order
     */
    get rows(): Row<Value, CellWithValue>[];
    /**
     * @param col The column coordinate
     * @returns The column
     */
    getColumn(col: number): Column<Value, CellWithValue>;
    /**
     * @returns All columns of the grid in ascending order
     */
    get columns(): Column<Value, CellWithValue>[];
    /**
     * @param callback A function that should be called, when a cell value of this grid changes
     * @returns a function to unregister the callback
     */
    onCellValueChanged(callback: (event: CellValueChangedEvent<Value>) => void): () => void;
}
//# sourceMappingURL=Grid.d.ts.map